#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <asthash.h>


/*--- ElfHash ---------------------------------------------------
 *  The published hash algorithm used in the UNIX ELF format
 *  for object files. Accepts a pointer to a string to be hashed
 *  and returns an unsigned long.
 *-------------------------------------------------------------*/
unsigned long ElfHash ( const unsigned char *name )
{
  unsigned long   h = 0, g;
  while ( *name )
    {
      h = ( h << 4 ) + *name++;
      if ( g = h & 0xF0000000 )
	h ^= g >> 24;
      h &= ~g;
    }
  return h;
}


struct string_hash_table *init_string_hash(char *name) {
  struct string_hash_table *newhash = malloc(sizeof(struct string_hash_table));
  memset(newhash,0,sizeof(struct string_hash_table));

  strncpy(newhash->name,name,sizeof(newhash->name));
  int key;
  for(key=0;key<STRING_HASH_TABLE_KEY_SIZE;key++) {
    newhash->data[key] = NULL;
  }

  return newhash;

}


int string_hash_write(struct string_hash_table *hash,char *name,char *value) {
  struct hash_storage *store,*last;
  unsigned long key = ElfHash(name) & STRING_HASH_TABLE_KEY_SIZE;
  store=last=NULL;

  store = hash->data[key];
  while(store) {
    last=store;
    store=store->next;
  }
  
  store = malloc(sizeof(struct hash_storage));
  memset(store,0,sizeof(struct hash_storage));
  store->var = malloc(strlen(name)+1);
  store->val = malloc(strlen(value)+1);
  strcpy(store->var,name);
  strcpy(store->val,value);
  
  if(last) 
    last->next = store;
  else 
    hash->data[key] = store;

  return 1;
}


char *string_hash_read(struct string_hash_table *hash,char *name) {
  unsigned long key = ElfHash(name) & STRING_HASH_TABLE_KEY_SIZE;
  struct hash_storage *store;

  store = hash->data[key];
  if(store) {

    if(! store->next)
      return store->val;

    else {
      store = store->next;
      while(store) {
	
	if(!strcmp(name,store->var)) {
	  return store->val;
	}
	store = store->next;
      }
    }
  }
  
  return NULL;
}


int string_hash_delete(struct string_hash_table *hash,char *name) {
  unsigned long key = ElfHash(name) & STRING_HASH_TABLE_KEY_SIZE;
  struct hash_storage *store=NULL,*last=NULL;
  int ret=0;
  
  
  store = hash->data[key];
  if(store) {
    if(! store->next) {
      hash->data[key] = store = NULL;
      free(store);
      ret=1;
    }
    else {
      store = store->next;
      
      while(store) {
	if(name && store->var && strcmp(store->var,name)) {
	  if(last)
	    last->next = store->next;
	  free(store);
	ret=1;
	}
	last = store;
      }
      
    }
  }
  
  return ret;
}


char *trim_spaces(char *var) {
  int y=0;
  

  for(;var[0] == ' ';var++);

  y = strlen(var)-1;
  for(y = strlen(var)-1;var[y] == ' ' ; y--) {
    (int) var[y] = NULL;
    (int) var[y+1] = ' ';
  }
  return var;
}

int pair2hash(struct string_hash_table *hash,char *pair) {
  char *var = pair;
  char *val = NULL;
  char *ret;
  char helper[50];
  int x=0,y=0;
  


  for(val=pair;val && y <= strlen(pair); val++ && y++) {
    if(val[0] == '=') {
      (int) val[0] = NULL;
      val++;
      x=1;
      break;
    }
  }


  if(x < 1) {
    val=pair;
    for(x=0;;x++) {
      sprintf(helper,"%d",x);
      ret = string_hash_read(hash,helper);
      if(!ret)
	break;
    }

    var = helper;
  }


  /* trim leading and trailing spaces */

  var=trim_spaces(var);
  val=trim_spaces(val);
  
  return  string_hash_write(hash,var,val);
}

void print_string_hash(struct string_hash_table *hash) {
  int key;
  struct hash_storage *store;

  for(key=0;key<STRING_HASH_TABLE_KEY_SIZE;key++) {
    if(hash->data[key]) {
      store = hash->data[key];

      if(store) {
	if(! store->next)
	  printf("'%s'='%s'\n",store->var,store->val);

	else {
	  store = store->next;
	  while(store) {
	    printf("'%s'='%s'\n",store->var,store->val);
	    store = store->next;
	  }
	}
      }

    }
  }
}

int named_args(struct string_hash_table *hash,char *data,char delim) {
  char *p,*args;
  args = data;
  int count=1;

  while((p = strchr(args,delim))) {
    *p = '\0';
    p++;

    pair2hash(hash,args);
    count++;
    args = p;
    p = NULL;
  }

  if(args && hash) 
    pair2hash(hash,args);
  
  return count;

}
